#define __PROG_TYPES_COMPAT__

#define F_CPU			16000000

#include <avr/pgmspace.h>
#include <avr/io.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>

#include "lcd.h"
#include "gfx.h"
#include "utils.h"

int note_frequencies[] = { 262, 294, 330, 349, 392, 440, 494, 523 };	// frecvente folosite pentru buzzer
static char game_table[DISPLAY_WIDTH][DISPLAY_HEIGHT];					// tabla de joc

// functie care face buzzerul sa "bazaie" :D
void make_it_buzz(int nota) {
	int period = 1000000 / note_frequencies[nota];
	int i;

	PORTD |= (1 << PD5);
	for (i = 0; i < period / 2; ++i)
		_delay_us(1);

	PORTD &= ~(1 << PD5);
	for (i = 0; i < period / 2; ++i)
		_delay_us(1);
}

// fiecare piesa are nevoie de niste limite pentru desenare si pentru coliziuni
struct bound {
	int16_t total_width, total_height;		// pentru a stii fiecare piesa cat de mare este
	uint16_t x_max, x_min, y_max, y_min;	// si de unde porneste
};

// fiecare piesa este facuta din cate 4 patratele
struct square {
	int16_t x, y;	// pozitia unui singur patratel
	uint16_t width;	// dimensiunea acestuia
	uint16_t color; // nu cred ca e nevoie de detalii :))
};

// detalii despre piesa si despre pozitia ei
struct options {
	uint8_t rotation;	// in ce cadran se afla 
	uint8_t type;		// tipul piesei
	uint8_t can_fall;	// piese mai poate cadea
};

// piesa in sine
struct piece {
	struct square elems[4]; // cele 4 patrate
	struct options opt;		// detalii legate de piesa
	struct bound s_bound;	// limitele piesei si pozitia in tabla
};

int8_t game_over = 0;	// jocul s-a terminat
int8_t lost = 0;			// cate jocuri a pierdut userul
int8_t last_button = UP; // by default se va selecta Play din meniu
int highscore_so_far;
int score;
int level = 1;			// nivelul de inceput
int level_speed = 250;	// reprezinta dificultatea nivelului, cu cat va fi mai mica cu atat mai rapid vor cadea piesele
int procent_level_speed = 30;

struct piece pieces[PIECES_NO];	// voi avea 4 piese disponibile, as putea sa mai adaug piesa Z dar pe viitor :)
struct piece square_piece;
struct piece bar_piece;
struct piece L_piece;
struct piece T_piece;

unsigned int i, j, k;

// verific daca un anumit buton a fost apasat 
// return 1 pentru true si 0 pentru false
static inline uint8_t is_button_pressed(uint8_t button)
{
	if (!(BUTTON_PIN & (1 << button))) {		
		if (!(BUTTON_PIN & (1 << button))) {			
			if (!(BUTTON_PIN & (1 << button))) {
				make_it_buzz(button);
				return 1;
			}
		}
	}
	
	return 0;
}

// mesajul de inceput de joc
static inline void welcome_user(int16_t x, int16_t y)
{
	LCD_fill_screen(ST7735_GREEN);
	
	GFX_set_text_size(2);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x+20, y);
	GFX_print_str("Hello!");

	GFX_set_text_size(1);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_BLUE);
	GFX_set_cursor(x+10, y + 50);
	GFX_print_str("#pm2017");
	
	GFX_set_text_color(ST7735_RED);
	GFX_set_cursor(x+10, y + 60);
	GFX_print_str("by cristighr");
}

// mesajul de sfarsit in momentul in care se da exit
static inline void put_byebye_on_that_lcd(int16_t x, int16_t y)
{
	LCD_fill_screen(ST7735_BLACK);
				
	GFX_set_text_size(2);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x+5, y);
	GFX_print_str("Good Bye");
	
	GFX_set_text_color(ST7735_YELLOW);
	GFX_set_cursor(x+5, y+20);
	GFX_print_str("GL & HF");
}

// mesajul cand se termina o runda de joc
// GAME OVER
static inline void put_game_over_on_that_lcd(int16_t x, int16_t y)
{	
	LCD_fill_screen(ST7735_RED);
	
	char your_score[20];

	GFX_set_text_size(2);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x+5, y);
	GFX_print_str("Game Over");

	GFX_set_text_size(1);
	GFX_set_rotation(1);

	sprintf(your_score, "Your score:  %d", score);
	GFX_set_text_color(ST7735_GREEN);
	GFX_set_cursor(x+10, y + 70);
	GFX_print_str(your_score);
}

// afiseaza meniul principal din joc
static inline void put_menu_on_that_lcd(int16_t x, int16_t y)
{
	LCD_fill_screen(ST7735_BLACK);
	
	char games_lost[10];
	char high_score[20];
	char level_string[20];

	GFX_set_text_size(2);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_BLUE);
	GFX_set_cursor(x+20, y);
	GFX_print_str("Tetris");

	GFX_set_text_size(1);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_YELLOW);
	GFX_set_cursor(x+10, y + 40);
	GFX_print_str("Play");

	GFX_set_text_color(ST7735_RED);
	GFX_set_cursor(x+10, y + 50);
	GFX_print_str("Exit");

	sprintf(games_lost, "Lost:  %d", lost);
	GFX_set_text_color(ST7735_GREEN);
	GFX_set_cursor(x+10, y + 70);
	GFX_print_str(games_lost);
	
	sprintf(high_score, "Highscore:  %d", highscore_so_far);
	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x+10, y + 80);
	GFX_print_str(high_score);
	
	sprintf(level_string, "Level:  %d", level);
	GFX_set_text_color(ST7735_MAGENTA);
	GFX_set_cursor(x+10, y + 60);
	GFX_print_str(level_string);
}

// in momentul in care se depaseste ultimul highscore se trece la alt nivel si se afiseaza o pagina noua
static inline void put_new_level_on_that_lcd(int16_t x, int16_t y)
{
	LCD_fill_screen(ST7735_BLUE);
	
	char level_string[20];

	GFX_set_text_size(2);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x+10, y);
	GFX_print_str("CONGRATS!");

	GFX_set_text_size(1);
	GFX_set_rotation(1);

	GFX_set_text_color(ST7735_YELLOW);
	GFX_set_cursor(x+2, y + 40);
	GFX_print_str("New level unlocked!");

	GFX_set_text_size(2);
	GFX_set_rotation(1);
	
	sprintf(level_string, "%d", level);
	GFX_set_text_color(ST7735_YELLOW);
	GFX_set_cursor(x+52, y + 50);
	GFX_print_str(level_string);
}

// initializarea pieselor 
// pozitiile lor, culoarea, rotatiile (pe 0) si cum arata
static inline void init_bar_piece()
{
	for (i = 0; i < PIECES_NO; ++i) {
		bar_piece.elems[i].width = SQUARE_DIM;
		bar_piece.elems[i].color = ST7735_CYAN;
		bar_piece.elems[i].x = X_MIN_UP;
	}
	
	bar_piece.elems[0].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 + 2 * SQUARE_DIM;
	bar_piece.elems[1].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 + SQUARE_DIM;
	bar_piece.elems[2].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2;
	bar_piece.elems[3].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 - SQUARE_DIM;
	
	bar_piece.s_bound.total_width = 4 * SQUARE_DIM;
	bar_piece.s_bound.total_height = SQUARE_DIM;
	bar_piece.s_bound.x_min = X_MIN_UP;
	bar_piece.s_bound.x_max = X_MIN_UP + bar_piece.s_bound.total_height;
	bar_piece.s_bound.y_min = bar_piece.elems[0].y - bar_piece.s_bound.total_width;
	bar_piece.s_bound.y_max = bar_piece.elems[0].y;
	
	bar_piece.opt.rotation = 0;
	bar_piece.opt.type = 0;
}

static inline void init_L_piece()
{
	for (i = 0; i < PIECES_NO; ++i) {
		L_piece.elems[i].width = SQUARE_DIM;
		L_piece.elems[i].color = ST7735_YELLOW;
		
		if (i != 0)
			L_piece.elems[i].x = X_MIN_UP + SQUARE_DIM;
		else 
			L_piece.elems[i].x = X_MIN_UP;
	}
	
	L_piece.elems[0].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 - SQUARE_DIM;
	L_piece.elems[1].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 + SQUARE_DIM;
	L_piece.elems[2].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2;
	L_piece.elems[3].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 - SQUARE_DIM;
	
	L_piece.s_bound.total_width = 3 * SQUARE_DIM;
	L_piece.s_bound.total_height = 2 * SQUARE_DIM;
	L_piece.s_bound.x_min = X_MIN_UP;
	L_piece.s_bound.x_max = X_MIN_UP + L_piece.s_bound.total_height;
	L_piece.s_bound.y_min = L_piece.elems[1].y - L_piece.s_bound.total_width;
	L_piece.s_bound.y_max = L_piece.elems[1].y;
	
	L_piece.opt.rotation = 0;
	L_piece.opt.type = 1;
	
}

static inline void init_T_piece()
{
	for (i = 0; i < PIECES_NO; i++) {
		T_piece.elems[i].width = SQUARE_DIM;
		T_piece.elems[i].color = ST7735_MAGENTA;
		
		if (i != 0)
			T_piece.elems[i].x = X_MIN_UP + SQUARE_DIM;
		else 
			T_piece.elems[i].x = X_MIN_UP;
	}
	
	T_piece.elems[0].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2;
	T_piece.elems[1].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 + SQUARE_DIM;
	T_piece.elems[2].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2;
	T_piece.elems[3].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 - SQUARE_DIM;
	
	T_piece.s_bound.total_width = 3 * SQUARE_DIM;
	T_piece.s_bound.total_height = 2 * SQUARE_DIM;
	T_piece.s_bound.x_min = X_MIN_UP;
	T_piece.s_bound.x_max = X_MIN_UP + T_piece.s_bound.total_height;
	T_piece.s_bound.y_min = T_piece.elems[1].y - T_piece.s_bound.total_width;
	T_piece.s_bound.y_max = T_piece.elems[1].y;
	
	T_piece.opt.rotation = 0;
	T_piece.opt.type = 2;
}

static inline void init_square_piece()
{
	for (i = 0; i < PIECES_NO; i++) {
		square_piece.elems[i].width = SQUARE_DIM;
		square_piece.elems[i].color = ST7735_BLUE;
		
		if (i <= 1)
			square_piece.elems[i].x = X_MIN_UP;
		else
			square_piece.elems[i].x = X_MIN_UP + SQUARE_DIM;
	}
	
	square_piece.elems[0].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 + SQUARE_DIM;
	square_piece.elems[1].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2;
	square_piece.elems[2].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2 + SQUARE_DIM;
	square_piece.elems[3].y = (Y_MIN_RIGHT + Y_MAX_LEFT) / 2;
	
	square_piece.s_bound.total_width = 2 * SQUARE_DIM;
	square_piece.s_bound.total_height = 2 * SQUARE_DIM;
	square_piece.s_bound.x_min = X_MIN_UP;
	square_piece.s_bound.x_max = X_MIN_UP + square_piece.s_bound.total_height;
	square_piece.s_bound.y_min = square_piece.elems[0].y - square_piece.s_bound.total_width;
	square_piece.s_bound.y_max = square_piece.elems[0].y;
	
	square_piece.opt.rotation = 0;
	square_piece.opt.type = 3;
}

static inline void init_pieces()
{
	init_bar_piece();
	init_L_piece();
	init_T_piece();
	init_square_piece();	

	pieces[0] = square_piece;
	pieces[1] = bar_piece;
	pieces[2] = L_piece;
	pieces[3] = T_piece;
}

// verific daca jocul s-a terminat
// am folosit functia pentru a fi mai usor de urmarit codul
static inline uint8_t is_game_over()
{
	return game_over;
}

// piesa mai poate cadea
static inline uint8_t piece_can_fall(struct piece s) 
{
	return s.opt.can_fall;
}

// verific daca pot sa desenez mai departe
static inline uint8_t can_draw_more(struct piece s)
{
	for (i = 0; i < PIECES_NO; i++)
		if (game_table[(s.s_bound.x_max + SQUARE_DIM - 1) / SQUARE_DIM][(s.elems[i].y - Y_OFFSET - 1) / SQUARE_DIM] == 1)
			return 0;
		
	return 1;
}

// desenez prima piesa
static inline void draw_initial_piece(struct piece &s) 
{
	// piesa poate cadea (nu a atins podeaua)
	s.opt.can_fall = 1;
	
	for (i = 0; i < PIECES_NO; i++) {
		// actualizez tabla de joc
		game_table[(s.elems[i].x - 1) / SQUARE_DIM][(s.elems[i].y - Y_OFFSET - 1) / SQUARE_DIM] = 1;
		
		GFX_draw_rect(s.elems[i].x,
					  s.elems[i].y, 
					  s.elems[i].width, // pentru ca desenez fiecare patratel in parte 
					  s.elems[i].width,	// si nu toata piesa deodata
					  s.elems[i].color, 
					  FILLED);			// vreau sa fie umplut
	}
}

static inline void clear_previous_piece(struct piece s)
{
	for (i = 0; i < PIECES_NO; i++) {
		// updatez tabla de joc
		game_table[(s.elems[i].x - 1) / SQUARE_DIM][(s.elems[i].y - Y_OFFSET - 1) / SQUARE_DIM] = 0;
		
		GFX_draw_rect(s.elems[i].x,
					  s.elems[i].y, 
					  s.elems[i].width, 
					  s.elems[i].width, 
					  ST7735_BLACK, // cu negru fac clear
					  FILLED);
	}
}

static inline void draw_next_piece(struct piece s)
{
	for (i = 0; i < PIECES_NO; i++) {
		game_table[(s.elems[i].x + SQUARE_DIM - 1) / SQUARE_DIM][(s.elems[i].y - Y_OFFSET - 1) / SQUARE_DIM] = 1;
		
		GFX_draw_rect(s.elems[i].x + SQUARE_DIM,
					  s.elems[i].y,
					  s.elems[i].width, 
					  s.elems[i].width, 
					  s.elems[i].color, 
					  FILLED);
	}
}

static inline void move_piece_left(struct piece &s)
{
	// daca ajunge la marginea din stanga sa nu iasa din ecran
	if (s.s_bound.y_max >= Y_MAX_LEFT - Y_OFFSET - 2*SQUARE_DIM)
		return; 
	
	// curat partea veche si o completez pe cea noua
	for (i = 0; i < PIECES_NO; i++) {
		// pozitia devine libera
		game_table[(s.elems[i].x + SQUARE_DIM - 1) / SQUARE_DIM][(s.elems[i].y - Y_OFFSET - 1) / SQUARE_DIM] = 0;				// cea veche
		game_table[(s.elems[i].x + SQUARE_DIM - 1) / SQUARE_DIM][(s.elems[i].y + SQUARE_DIM - Y_OFFSET - 1) / SQUARE_DIM] = 1;	// cea noua
		
		GFX_draw_rect(s.elems[i].x + SQUARE_DIM,
					  s.elems[i].y,
					  s.elems[i].width, 
					  s.elems[i].width, 
					  ST7735_BLACK, 
					  FILLED);		// cea veche
					  
		GFX_draw_rect(s.elems[i].x + SQUARE_DIM,
					  s.elems[i].y + SQUARE_DIM,
					  s.elems[i].width, 
					  s.elems[i].width, 
					  s.elems[i].color, 
					  FILLED);		// cea noua
	}
	
	// actualizez pozitiile noi
	for (i = 0; i < PIECES_NO; i++) 
		s.elems[i].y += SQUARE_DIM;
	
	s.s_bound.y_max += SQUARE_DIM;
	s.s_bound.y_min += SQUARE_DIM;
}

static inline void move_piece_right(struct piece &s)
{
	// verific sa nu iasa din ecran in partea dreapta
	if (s.s_bound.y_min <= Y_MIN_RIGHT - Y_OFFSET + 3)
		return;
	
	for (i = 0; i < PIECES_NO; i++) {
		game_table[(s.elems[i].x + SQUARE_DIM - 1) / SQUARE_DIM][(s.elems[i].y - Y_OFFSET - 1) / SQUARE_DIM] = 0;				// cea veche
		game_table[(s.elems[i].x + SQUARE_DIM - 1) / SQUARE_DIM][(s.elems[i].y - SQUARE_DIM - Y_OFFSET - 1) / SQUARE_DIM] = 1;	// cea noua
		
		GFX_draw_rect(s.elems[i].x + SQUARE_DIM,
					  s.elems[i].y,
					  s.elems[i].width, 
					  s.elems[i].width, 
					  ST7735_BLACK, 
					  FILLED);	// veche
		
		GFX_draw_rect(s.elems[i].x + SQUARE_DIM,
					  s.elems[i].y - SQUARE_DIM,
					  s.elems[i].width, 
					  s.elems[i].width, 
					  s.elems[i].color, 
					  FILLED);	// noua
		
	}
	
	for (i = 0; i < 4; i++)
		s.elems[i].y -= SQUARE_DIM;

	s.s_bound.y_max -= SQUARE_DIM;
	s.s_bound.y_min -= SQUARE_DIM;
}

// verific ce butoane au fost apasate in timpul jocului 
static inline void check_buttons_if_pressed(struct piece &s)
{
	if (is_button_pressed(BUTTON_LEFT))
			move_piece_left(s);
		
	if (is_button_pressed(BUTTON_RIGHT))
		move_piece_right(s);

	if (is_button_pressed(BUTTON_ROTATE))
		s.opt.rotation++;
			
	if (is_button_pressed(BUTTON_SURPRISE))
		s.opt.rotation--;
}

// modific coordonatele in cazul caderii piesei
static inline void update_falling_positions(struct piece &s)
{
	for (i = 0; i < PIECES_NO; i++)
		s.elems[i].x += SQUARE_DIM;
	
	s.s_bound.x_max += SQUARE_DIM;
	s.s_bound.x_min += SQUARE_DIM;
}

// verific daca s-a ajuns la marginea de jos
static void check_if_reached_bottom(struct piece &s)
{
	if (s.s_bound.x_max >= X_MAX_DOWN) {
		s.s_bound.x_max = X_MAX_DOWN;
		s.opt.can_fall = 0;
	}	
}

// rotatiile pieselor
// antetul functiilor este in felul urmator
// rotate_<numarul cadranului>_type_<tipul piesei>
// 0 reprezinta rotatia de baza (by default)
// urmeaza 1 la dreapta in sensul acelor de ceasornic
static inline void rotate_zero_type_zero(struct piece &s)
{
	for (i = 1; i < PIECES_NO; i++) {
		s.elems[i].y = s.elems[0].y - i * SQUARE_DIM;
		s.elems[i].x = s.elems[0].x;
	}
	
	s.s_bound.total_width = 4 * SQUARE_DIM;
	s.s_bound.total_height = SQUARE_DIM;
	
	s.s_bound.x_min = s.elems[0].x;
	s.s_bound.x_max = s.elems[0].x + s.s_bound.total_height;
	
	s.s_bound.y_min = s.elems[0].y - s.s_bound.total_width;
	s.s_bound.y_max = s.elems[0].y;
}

static inline void rotate_zero_type_one(struct piece &s)
{
	for (i = 1; i < PIECES_NO; i++)
		s.elems[i].x = s.elems[0].x + SQUARE_DIM;
		
	s.elems[1].y = s.elems[0].y + 2 * SQUARE_DIM;
	s.elems[2].y = s.elems[0].y + SQUARE_DIM;
	s.elems[3].y = s.elems[0].y;
	
	s.s_bound.total_width = 3 * SQUARE_DIM;
	s.s_bound.total_height = 2 * SQUARE_DIM;
	
	s.s_bound.x_min = s.elems[0].x;
	s.s_bound.x_max = s.elems[0].x + s.s_bound.total_height;
	
	s.s_bound.y_min = s.elems[1].y - s.s_bound.total_width;
	s.s_bound.y_max = s.elems[1].y;
}

static inline void rotate_zero_type_two(struct piece &s)
{
	for (i = 1; i < PIECES_NO; i++)
		s.elems[i].x = s.elems[0].x + SQUARE_DIM;
		
	s.elems[1].y = s.elems[0].y + SQUARE_DIM;
	s.elems[2].y = s.elems[0].y;
	s.elems[3].y = s.elems[0].y - SQUARE_DIM;
	
	s.s_bound.total_width = 3 * SQUARE_DIM;
	s.s_bound.total_height = 2 * SQUARE_DIM;
	
	s.s_bound.x_min = s.elems[0].x;
	s.s_bound.x_max = s.elems[0].x + s.s_bound.total_height;
	
	s.s_bound.y_min = s.elems[1].y - s.s_bound.total_width;
	s.s_bound.y_max = s.elems[1].y;
}

static inline void apply_rotation_zero(struct piece &s)
{
	if (s.opt.rotation % 4 == 0) {
		if (s.opt.type == 0)
			rotate_zero_type_zero(s);
		
		if (s.opt.type == 1)
			rotate_zero_type_one(s);
		
		if (s.opt.type == 2) 
			rotate_zero_type_two(s);
	}
}

static inline void rotate_one_type_zero(struct piece &s)
{
	for (i = 1; i < 4; i++) {
		s.elems[i].x = s.elems[0].x + i * SQUARE_DIM;
		s.elems[i].y = s.elems[0].y;
	}
	
	s.s_bound.total_width = SQUARE_DIM;
	s.s_bound.total_height = 4 * SQUARE_DIM;
	
	s.s_bound.x_min = s.elems[0].x;
	s.s_bound.x_max = s.elems[0].x + s.s_bound.total_height;
	
	s.s_bound.y_min = s.elems[0].y - s.s_bound.total_width;
	s.s_bound.y_max = s.elems[0].y;
}

static inline void rotate_one_type_one(struct piece &s)
{
	s.elems[0].x = s.elems[3].x;
	s.elems[1].x = s.elems[3].x - 2 * SQUARE_DIM;
	s.elems[2].x = s.elems[3].x - SQUARE_DIM;
	s.elems[1].y = s.elems[2].y = s.elems[3].y;
	s.elems[0].y = s.elems[3].y - SQUARE_DIM;
	
	s.s_bound.total_width = 2 * SQUARE_DIM;
	s.s_bound.total_height = 3 * SQUARE_DIM;
	
	s.s_bound.x_min = s.elems[1].x;
	s.s_bound.x_max = s.elems[1].x + s.s_bound.total_height;
	
	s.s_bound.y_min = s.elems[1].y - s.s_bound.total_width;
	s.s_bound.y_max = s.elems[1].y;
}

static inline void rotate_one_type_two(struct piece &s)
{
	s.elems[1].y = s.elems[3].y = s.elems[2].y;
	s.elems[0].y = s.elems[2].y - SQUARE_DIM;
	s.elems[0].x = s.elems[2].x;
	s.elems[1].x = s.elems[2].x - SQUARE_DIM;
	s.elems[3].x = s.elems[2].x + SQUARE_DIM;
	
	s.s_bound.total_width = 2 * SQUARE_DIM;
	s.s_bound.total_height = 3 * SQUARE_DIM;
	
	s.s_bound.x_min = s.elems[1].x;
	s.s_bound.x_max = s.elems[1].x + s.s_bound.total_height;
	
	s.s_bound.y_min = s.elems[1].y - s.s_bound.total_width;
	s.s_bound.y_max = s.elems[1].y;
}

static inline void apply_rotation_one(struct piece &s)
{
	if (s.opt.rotation % 4 == 1 || s.opt.rotation % 4 == -3) {
		if (s.opt.type == 0)
			rotate_one_type_zero(s);
		
		if (s.opt.type == 1)
			rotate_one_type_one(s);
		
		if (s.opt.type == 2)
			rotate_one_type_two(s);
	}
}

static inline void rotate_two_type_zero(struct piece &s)
{
	for (i = 1; i < 4; i++) {
		s.elems[i].y = s.elems[0].y + i * SQUARE_DIM;
		s.elems[i].x = s.elems[0].x;
	}
	
	s.s_bound.total_width = 4 * SQUARE_DIM;
	s.s_bound.total_height = SQUARE_DIM;
	
	s.s_bound.x_min = s.elems[0].x;
	s.s_bound.x_max = s.elems[0].x + s.s_bound.total_height;
	
	s.s_bound.y_min = s.elems[0].y - s.s_bound.total_width;
	s.s_bound.y_max = s.elems[0].y;
}

static inline void rotate_two_type_one(struct piece &s)
{
	s.elems[1].x = s.elems[2].x = s.elems[3].x;
	s.elems[0].x = s.elems[3].x + SQUARE_DIM;
	s.elems[0].y = s.elems[3].y;
	s.elems[2].y = s.elems[3].y - SQUARE_DIM;
	s.elems[1].y = s.elems[3].y - 2 * SQUARE_DIM;
	
	s.s_bound.total_width = 3 * SQUARE_DIM;
	s.s_bound.total_height = 2 * SQUARE_DIM;
	
	s.s_bound.x_min = s.elems[3].x;
	s.s_bound.x_max = s.elems[3].x + s.s_bound.total_height;
	
	s.s_bound.y_min = s.elems[3].y - s.s_bound.total_width;
	s.s_bound.y_max = s.elems[3].y;
}

static inline void rotate_two_type_two(struct piece &s)
{
	s.elems[1].x = s.elems[3].x = s.elems[2].x;
	s.elems[0].x = s.elems[2].x + SQUARE_DIM;
	s.elems[1].y = s.elems[2].y - SQUARE_DIM;
	s.elems[3].y = s.elems[2].y + SQUARE_DIM;
	s.elems[0].y = s.elems[2].y;
	
	s.s_bound.total_width = 3 * SQUARE_DIM;
	s.s_bound.total_height = 2 * SQUARE_DIM;
	
	s.s_bound.x_min = s.elems[1].x;
	s.s_bound.x_max = s.elems[1].x + s.s_bound.total_height;
	
	s.s_bound.y_min = s.elems[3].y - s.s_bound.total_width;
	s.s_bound.y_max = s.elems[3].y;
}

static inline void apply_rotation_two(struct piece &s)
{
	if (s.opt.rotation % 4 == 2 || s.opt.rotation % 4 == -2) {
		if (s.opt.type == 0)
			rotate_two_type_zero(s);
		
		if (s.opt.type == 1)
			rotate_two_type_one(s);
		
		if (s.opt.type == 2)
			rotate_two_type_two(s);
	}
}

static inline void rotate_three_type_zero(struct piece &s)
{
	for (i = 1; i < 4; i++) {
		s.elems[i].x = s.elems[0].x - i * SQUARE_DIM;
		s.elems[i].y = s.elems[0].y;
	}
		
	s.s_bound.total_width = SQUARE_DIM;
	s.s_bound.total_height = 4 * SQUARE_DIM;
	
	s.s_bound.x_min = s.elems[0].x;
	s.s_bound.x_max = s.elems[0].x + s.s_bound.total_height;
	
	s.s_bound.y_min = s.elems[0].y - s.s_bound.total_width;
	s.s_bound.y_max = s.elems[0].y;
}

static inline void rotate_three_type_one(struct piece &s)
{
	s.elems[0].x = s.elems[3].x;
	s.elems[1].x = s.elems[3].x + 2 * SQUARE_DIM;
	s.elems[2].x = s.elems[3].x + SQUARE_DIM;
	s.elems[1].y = s.elems[2].y = s.elems[3].y;
	s.elems[0].y = s.elems[3].y + SQUARE_DIM;
	
	s.s_bound.total_width = 2 * SQUARE_DIM;
	s.s_bound.total_height = 3 * SQUARE_DIM;
	
	s.s_bound.x_min = s.elems[3].x;
	s.s_bound.x_max = s.elems[3].x + s.s_bound.total_height;
	
	s.s_bound.y_min = s.elems[0].y - s.s_bound.total_width;
	s.s_bound.y_max = s.elems[0].y;
}

static inline void rotate_three_type_two(struct piece &s)
{
	s.elems[1].y = s.elems[3].y = s.elems[2].y;
	s.elems[0].y = s.elems[2].y + SQUARE_DIM;
	s.elems[0].x = s.elems[2].x;
	s.elems[1].x = s.elems[2].x + SQUARE_DIM;
	s.elems[3].x = s.elems[2].x - SQUARE_DIM;
	
	s.s_bound.total_width = 2 * SQUARE_DIM;
	s.s_bound.total_height = 3 * SQUARE_DIM;
	
	s.s_bound.x_min = s.elems[3].x;
	s.s_bound.x_max = s.elems[3].x + s.s_bound.total_height;
	
	s.s_bound.y_min = s.elems[0].y - s.s_bound.total_width;
	s.s_bound.y_max = s.elems[0].y;
}

static inline void apply_rotation_three(struct piece &s)
{
	if (s.opt.rotation % 4 == 3 || s.opt.rotation % 4 == -1) {
		if (s.opt.type == 0)
			rotate_three_type_zero(s);
		
		if (s.opt.type == 1)
			rotate_three_type_one(s);
		
		if (s.opt.type == 2)
			rotate_three_type_two(s);
	}
}

static inline void apply_rotations_if_any(struct piece &s)
{
	apply_rotation_zero(s);
	apply_rotation_one(s);
	apply_rotation_two(s);
	apply_rotation_three(s);
}

static inline void check_if_game_over(struct piece s)
{
	if (s.s_bound.x_min <= X_MIN_UP + SQUARE_DIM) {
		game_over = 1;
		lost++;
	}
}

// logica din momentul in care o piesa apare pe ecran pana nu mai are unde sa cada
static void put_piece_on_that_lcd(struct piece s)
{
	// verific starea jocului
	if (is_game_over())
		return;

	// deseneaza prima piesa
	draw_initial_piece(s);
	
	// daca piese mai poate cadea si nu s-a terminat jocul
	while (piece_can_fall(s) && !is_game_over()) {

		// verific daca pot sa desenez la urmatoarea iteratie
		s.opt.can_fall = can_draw_more(s);

		// sterg piesa veche
		clear_previous_piece(s);

		// aplic rotatiile necesare in cazul in care am vrut o rotatie
		apply_rotations_if_any(s);

		// desenez pe ecran piesa
		draw_next_piece(s);

		// pot sa modific delay-ul pentru nivel de exemplu 
		// cu cat e mai mic cu atat dificultatea e mai mare :>
		_delay_ms(level_speed);

		// verific pe ce butoane a apasat userul
		// daca a apasat vreunul...
		check_buttons_if_pressed(s);

		// actualizez pozitia piese in timp ce cade
		update_falling_positions(s);
		
		// daca a ajuns la podea ma opresc
		check_if_reached_bottom(s);
	}

	check_if_game_over(s);
}

static void init_buttons()
{
	BUTTON_REG &= ~((1 << BUTTON_LEFT) | (1 << BUTTON_RIGHT) | (1 << BUTTON_ROTATE) | (1 << BUTTON_SURPRISE));
	BUTTON_PORT |= ((1 << BUTTON_LEFT) | (1 << BUTTON_RIGHT) | (1 << BUTTON_ROTATE) | (1 << BUTTON_SURPRISE));
}

static void init_buzzer()
{
	BUZZER_REG |= (1 << PD5);
	BUZZER_PORT &= ~(1 << PD5);
}

static inline void init_game_table()
{
	for (i = 0; i < DISPLAY_HEIGHT; i++)
		for (j = 0; j < DISPLAY_WIDTH; j++)
			game_table[i][j] = 0;
}

static inline void menu_buttons_update()
{
	if (last_button == UP)
		GFX_draw_rect(68, 88, 10, 28, ST7735_CYAN, 0);
	else if (last_button == DOWN)
		GFX_draw_rect(78, 88, 10, 28, ST7735_CYAN, 0);

	if (is_button_pressed(BUTTON_SURPRISE)) {
		last_button = DOWN;
		GFX_draw_rect(68, 88, 10, 28, ST7735_BLACK, 0); // sterg vechiul patrat
		GFX_draw_rect(78, 88, 10, 28, ST7735_CYAN, 0);	// il pun pe urmatorul
	}
	else if (is_button_pressed(BUTTON_ROTATE)) {
		last_button = UP;
		GFX_draw_rect(78, 88, 10, 28, ST7735_BLACK, 0); // sterg vechiul patrat
		GFX_draw_rect(68, 88, 10, 28, ST7735_CYAN, 0);	// il pun pe urmatorul
	}
}

static inline uint8_t is_play_selected()
{
	if (last_button == UP && is_button_pressed(BUTTON_LEFT))
		return 1;
		
	return 0;
}

static inline uint8_t is_exit_selected()
{
	if (last_button == DOWN && is_button_pressed(BUTTON_LEFT))
		return 1;
	
	return 0;
}

static void put_score_on_that_lcd(int16_t x, int16_t y)
{			
	char current_score[20];
	
	GFX_set_text_size(1);
	GFX_set_rotation(1);

	sprintf(current_score, "score: %d", score);
	GFX_set_text_color(ST7735_WHITE);
	GFX_set_cursor(x, y);
	GFX_print_str("");
	GFX_print_str(current_score);
}

static inline int get_random_position(int min, int max) 
{
	return ((rand() % (max + 1 - min)) + min);
}

// bucla jocului
// de aici magic will happen
static void main_game_loop()
{
	while (!is_game_over()) {
		put_score_on_that_lcd(70, 5);
		
		// determin o piesa random
		int random_index = ((uint8_t) rand()) % PIECES_NO;
		
		// variez de unde va "cadea" piesa 
		int modifier = get_random_position(0, 10);
		
		for (i = 0; i < PIECES_NO; ++i)
			pieces[random_index].elems[i].y -= modifier;
		
		pieces[random_index].s_bound.y_min -= modifier;
		pieces[random_index].s_bound.y_max -= modifier;	
			
		put_piece_on_that_lcd(pieces[random_index]);
		score++;
	}		
}

// am compus niste muzica pentru inceputul jocului si pentru new level unlocked
static inline void play_music_begin_game()
{	
	for (i = 0; i < 80; i++)
		make_it_buzz(4);
		
	_delay_us(4);
	
	for (i = 0; i < 20; i++)
		make_it_buzz(2);
	
	_delay_us(4);
	
	for (i = 0; i < 80; i++)
		make_it_buzz(4);
		
	_delay_us(4);
	
	for (i = 0; i < 20; i++)
		make_it_buzz(2);
		
	_delay_us(4);
	
	for (i = 0; i < 80; i++)
		make_it_buzz(0);
		
	for (i = 0; i < 80; i++)
		make_it_buzz(7);
}

static inline void play_music_unlock_level()
{
	for (j = 0; j < 3; j++) {
		for (i = 0; i < 50; i++)
			make_it_buzz(5);
		
		_delay_us(2);
	
		for (i = 0; i < 50; i++)
			make_it_buzz(2);
	}			
		
	for (j = 0; j < 3; j++) {
		for (i = 0; i < 50; i++)
			make_it_buzz(6);
			
		_delay_us(2);
		
		for (i = 0; i < 50; i++)
			make_it_buzz(7);
	}			
}

// refac startul
// resetez variabilele
static inline void refresh()
{
	game_over = 0;
	last_button = UP;
	init_pieces();
	init_game_table();
	
	if (score > highscore_so_far) {
		highscore_so_far = score;
		level++;	// cresc nivelul si o sa creasca si dificultatea, piesele or sa cada mai repede
		level_speed -= procent_level_speed;
		
		put_new_level_on_that_lcd(5, 30);
		
		play_music_unlock_level();
		
		_delay_ms(3000);
		
		LCD_fill_screen(ST7735_BLACK);
	}
	
	score = 0;
}

static void play_game()
{
	// incepe jocul 
	main_game_loop();

	if (is_game_over()) {
		put_game_over_on_that_lcd(5, 30);
		
		_delay_ms(2000);
		
		refresh();
		
		LCD_fill_screen(ST7735_BLACK);
		put_menu_on_that_lcd(5, 30);
	}
}

int main(void)
{
	init_buttons();
	init_buzzer();

	init_pieces();
	init_game_table();

	LCD_init();
	GFX_init(128, 128);
	
	welcome_user(5, 30);
	
	_delay_ms(3000);
	
	put_menu_on_that_lcd(5, 30);
	
	play_music_begin_game();
	
	for(;;) {
		menu_buttons_update(); // fac update la meniu in cazul in care se apasa butoane

		if (is_play_selected()) {
			LCD_fill_screen(ST7735_BLACK);
			_delay_ms(1000);

			play_game();
		}

		if (is_exit_selected()) {
			put_byebye_on_that_lcd(5, 30);
			_delay_ms(3000);
			LCD_fill_screen(ST7735_BLACK);
			LCD_end();
			break;
		}
	}

	return 0;
}
